#!/usr/bin/env python3.1

# Copyright (C) 2012  Internet Systems Consortium.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SYSTEMS CONSORTIUM
# DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# INTERNET SYSTEMS CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import socket, struct, sys
from isc.dns import *

# The IPv4 address of f.root-servers.net (of the day), which is known to
# respond to AXFR queries for the root zone.
DEFAULT_ROOT_SERVER_ADDRESS = '192.5.5.241'

# These are the types of TLDs in terms of this test.  The "descriptions"
# table will be used to generate the final output.
TLDTYPE_UNKNOWN = 0
TLDTYPE_UNSIGNED = 1
TLDTYPE_NSEC_SIGNED = 2
TLDTYPE_NSEC3_SIGNED = 3
TLDTYPE_NSEC3_OPTOUT_SIGNED = 4
TLDTYPE_NSEC3_UNKNOWN = 5
tldtype_descriptions = { TLDTYPE_UNKNOWN : 'unknown',
                         TLDTYPE_UNSIGNED : 'unsigned',
                         TLDTYPE_NSEC_SIGNED : 'signed with NSEC',
                         TLDTYPE_NSEC3_SIGNED : 'signed with NSEC3, no optout',
                         TLDTYPE_NSEC3_OPTOUT_SIGNED :
                             'signed with NSEC3, optout',
                         TLDTYPE_NSEC3_UNKNOWN :
                             'signed with NSEC3, may/maynot optout' }

# Store statistics about NSEC3 iterations here (this is defined as global
# just for convenience)
iteration_stats = {}

# Fetch the root zone data from a root server that allows AXFR, and retrieve
# top level domains and their glue address records.
def get_tlds(root_server_addr, tlds, v6_glues, v4_glues):
    # Build AXFR query
    axfr_query = Message(Message.RENDER)
    axfr_query.set_opcode(Opcode.QUERY())
    axfr_query.set_rcode(Rcode.NOERROR())
    axfr_query.add_question(Question(Name('.'), RRClass.IN(), RRType.AXFR()))
    renderer = MessageRenderer()
    axfr_query.to_wire(renderer)

    # Connect to the specified root server and send the query.
    # Since this is TCP, we first need to send 2-byte length data.
    gai = socket.getaddrinfo(root_server_addr, 53, socket.AF_UNSPEC,
                             socket.SOCK_STREAM, 0,
                             socket.AI_NUMERICHOST|socket.AI_NUMERICSERV)[0]
    s = socket.socket(gai[0], gai[1], gai[2])
    s.connect(gai[4])
    s.send(struct.pack('H', socket.htons(renderer.get_length())))
    s.send(renderer.get_data())

    # Get the response until we see 2 SOAs.  Look for all non-root NS
    # records, and all AAAA and A records, and remember them in the appropriate
    # context.
    axfr_response = Message(Message.PARSE)
    soa_count = 0
    while soa_count < 2:
        recv_data = s.recv(2, socket.MSG_WAITALL)
        msg_len = socket.htons(struct.unpack('H', recv_data)[0])
        rcvd_data = s.recv(msg_len, socket.MSG_WAITALL)

        axfr_response.clear(Message.PARSE)
        axfr_response.from_wire(rcvd_data, Message.PRESERVE_ORDER)
        for rr in axfr_response.get_section(Message.SECTION_ANSWER):
            if rr.get_type() == RRType.SOA():
                soa_count += 1
                continue
            owner_txt = rr.get_name().to_text().lower()
            if rr.get_type() == RRType.NS() and owner_txt != '.':
                if owner_txt not in tlds:
                    tlds[owner_txt] = []
                tlds[owner_txt].append(rr.get_rdata()[0].to_text())
            elif rr.get_type() == RRType.AAAA():
                if owner_txt not in v6_glues:
                    v6_glues[owner_txt] = []
                v6_glues[owner_txt].append(rr.get_rdata()[0].to_text())
            elif rr.get_type() == RRType.A():
                if owner_txt not in v4_glues:
                    v4_glues[owner_txt] = []
                v4_glues[owner_txt].append(rr.get_rdata()[0].to_text())

# A helper function that resolves a given pair of domain name and RR type
# from any of the given IPv6 and IPv4 addresses.  If must_be_auth is True,
# it only looks for authoritative answers.
# Return the response (a Message object) that is first found or None if
# all attempts fail.
def do_resolve(name, type, v6addrs, v4addrs, must_be_auth=True):
    # Build the query message, with DNSSEC DO bit on, UDP buf size of 4096
    query = Message(Message.RENDER)
    query.set_opcode(Opcode.QUERY())
    query.set_rcode(Rcode.NOERROR())
    query.add_question(Question(name, RRClass.IN(), type))
    edns = EDNS()
    edns.set_dnssec_awareness(True)
    edns.set_udp_size(4096)
    query.set_edns(edns)

    # Render the message to wire-format data
    query_renderer = MessageRenderer()
    query.to_wire(query_renderer)

    # Try all given addresses until we get an expected response.  Realistically
    # the older version of IP addresses are still more reachable today, so we
    # try them first.
    response = Message(Message.PARSE)
    for addr in v4addrs + v6addrs:
        gai = socket.getaddrinfo(addr, 53, socket.AF_UNSPEC, socket.SOCK_DGRAM,
                                 0,
                                 socket.AI_NUMERICHOST|socket.AI_NUMERICSERV)[0]
        s = socket.socket(gai[0], gai[1], gai[2])
        s.sendto(query_renderer.get_data(), gai[4])

        response.clear(Message.PARSE)
        s.settimeout(10)        # don't wait too long
        try:
            response.from_wire(s.recvfrom(4096)[0])
        except socket.timeout:  # consider it unreachable, try the next one.
            continue
        if must_be_auth and \
                not response.get_header_flag(Message.HEADERFLAG_AA):
            sys.stderr.write('%s returned non authoritative data for %s\n' %
                             (addr, name.to_text()))
            continue
        return response
    return None

# See if a given type of RR(set) exists in the given section of the message.
def find_rrset(type, msg, section):
    for rrset in msg.get_section(section):
        if rrset.get_type() == type:
            return rrset
    return None

# Check if a TLD (that is known to be signed with NSEC3) uses opt-out
# NSEC3.
def check_nsec3(tld, v6addrs, v4addrs):
    global iteration_stats

    probe_names = [prefix + tld for prefix in
                   ['google.', 'google.com.', 'google.co.', 'aaa']]
    for probe_name in probe_names:
        response = do_resolve(Name(probe_name), RRType.SOA(), v6addrs, v4addrs,
                              False)
        if response is None:
            continue
        # if it's not indicate delegation, ignore it.
        if response.get_header_flag(Message.HEADERFLAG_AA) or \
                find_rrset(RRType.NS(), response,
                           Message.SECTION_AUTHORITY) is None:
            continue
        if find_rrset(RRType.DS(), response,
                      Message.SECTION_AUTHORITY) is not None:
            # This issecure delegation
            continue
        nsec3 = find_rrset(RRType.NSEC3(), response, Message.SECTION_AUTHORITY)
        if nsec3 is None:
            sys.stderr.write('Broken NSEC3 signed zone: %s\n' % tld)
            continue
        nsec3_fields = nsec3.get_rdata()[0].to_text().split()
        if int(nsec3_fields[0]) != 1:
            sys.stderr.write('%s uses an unknown NSEC3 algorithm: %s\n' %
                             (tld, nsec3_fields[0]))
        iteration = int(nsec3_fields[2])
        if iteration not in iteration_stats:
            iteration_stats[iteration] = []
        iteration_stats[iteration].append(tld)
        if (int(nsec3_fields[1]) & 1) == 0:
            return TLDTYPE_NSEC3_SIGNED
        return TLDTYPE_NSEC3_OPTOUT_SIGNED
    return TLDTYPE_NSEC3_UNKNOWN

# Check the DNSSEC type of the given TLD with the known NS addresses.
# Sometimes it tries some child zones of the given TLD (see below), and
# to signal such cases to the caller, it returns the actually used zone name
# as well as the identified type.
def check_dnssec(tld, v6addrs, v4addrs):
    response = do_resolve(Name(tld), RRType.NSEC(), v6addrs, v4addrs)
    if response is None:
        sys.stderr.write('No information is available for %s\n' % tld)
        return TLDTYPE_UNKNOWN, tld
    # If the answer section isn't empty, the zone should be signed with NSEC.
    if response.get_rr_count(Message.SECTION_ANSWER) > 0:
        # Apply heuristic: sometimes com.$tld and co.$tld exist as separate
        # zones, which may be signed with NSEC3.  In that case we should
        # rather check them.  For simplicity we assume the same servers have
        # the authority for such child zones.
        for sub in ['com.', 'co.']:
            response = do_resolve(Name(sub + tld), RRType.NSEC(), v6addrs,
                                  v4addrs)
            if response is not None and \
                    response.get_rr_count(Message.SECTION_ANSWER) == 0 and \
                    find_rrset(RRType.NSEC3(), response,
                               Message.SECTION_AUTHORITY) is not None:
                return check_nsec3(sub + tld, v6addrs, v4addrs), sub + tld
        return TLDTYPE_NSEC_SIGNED, tld
    # Otherwise, check if the authority section has NSEC3.  If it does, the
    # zone is signed with NSEC3; otherwise, it's unsigned.
    if find_rrset(RRType.NSEC3(), response,
                  Message.SECTION_AUTHORITY) is not None:
        return check_nsec3(tld, v6addrs, v4addrs), tld
    return TLDTYPE_UNSIGNED, tld

def probe_tld(tld, nameservers, v6_glues, v4_glues, results):
    # get the known IPv6 and IPv4 address of the TLD's NS
    v6addrs = []
    v4addrs = []
    for ns in nameservers:
        if ns in v6_glues:
            v6addrs.extend(v6_glues[ns])
        if ns in v4_glues:
            v4addrs.extend(v4_glues[ns])
    if len(v4addrs) == 0 and len(v6addrs) == 0:
        sys.stderr.write('No NS addresses are available for %s\n' % tld)
        return
    check_result, checked_tld = check_dnssec(tld, v6addrs, v4addrs)
    results[check_result].append(checked_tld)

if __name__ == '__main__':
    tlds = {}
    v6_glues = {}
    v4_glues = {}
    get_tlds(DEFAULT_ROOT_SERVER_ADDRESS, tlds, v6_glues, v4_glues)
    sys.stdout.write('found %d TLDs, %d IPv6 addresses, %d IPv4 addresses\n' %
                     (len(tlds), len(v6_glues), len(v4_glues)))
    probe_results = [[] for i in range(0, TLDTYPE_NSEC3_UNKNOWN + 1)]
    for (tld, nameservers) in tlds.items():
        probe_tld(tld, nameservers, v6_glues, v4_glues, probe_results)
    for i in range(0, TLDTYPE_NSEC3_UNKNOWN + 1):
        sys.stdout.write('%s: %d' % (tldtype_descriptions[i],
                                     len(probe_results[i])))
        sys.stdout.write(' (' + ', '.join(sorted(probe_results[i])) + ')')
        sys.stdout.write('\n')
    sys.stdout.write('\nStatistics about NSEC3 iterations:\n')
    for it in sorted(iteration_stats.keys()):
        sys.stdout.write('%d: %d' % (it, len(iteration_stats[it])))
        sys.stdout.write(' (' + ', '.join(sorted(iteration_stats[it])) + ')')
        sys.stdout.write('\n')
